// render.c
// (C) 1998 lars brinkhoff.

#define X_MIN -100.0
#define X_MAX 100.0
#define Y_MIN -100.0
#define Y_MAX 100.0

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "ray.h"
#include "lasr.h"

static Real render_pixel (Real x, Real y, Real a, Real c, int n)
{
  Real b, d, u, v;
  Real I = 0;
  int i, j;
  Ray ray;
  
  b = a / n;
  d = c / n;
  
  ray.p = vec (0.0, 0.0, -150.0);
  ray.depth = 0;

  for (i = 0, u = 0.0; i < n; i++, u += b)
    for (j = 0, v = 0.0; j < n; j++, v += d)
      {
	ray.d.x = x - ray.p.x + u + myrand (-0.5 * b, 0.5 * b, 1);
	ray.d.y = y - ray.p.y + v + myrand (-0.5 * d, 0.5 * d, 1);
	ray.d.z = 0.0 - ray.p.z;
	vector_normalize (ray.d);

	ray.t = myrand (0, 1, 1);
	
	I += trace (ray, i, j, n);
      }
  
  return I / (n * n);
}

void
render_area (int i0, int j0, int w, int h, int view_width, int view_height,
	     int n)
{
  Real intensity0[w][h], intensity1[w][h];
  Real delta, deltamax = 0, sum = 0;
  Real x, y, delta_x, delta_y;
  int i, j;
  Real I;
  int J;
  int k = 0;
  
  delta_x = (X_MAX - X_MIN) / view_width;
  delta_y = (Y_MAX - Y_MIN) / view_height;
  
  for (j = 0, y = Y_MAX - j0 * delta_y; j < h; j++, y -= delta_y)
    for (i = 0, x = X_MIN + i0 * delta_x; i < w; i++, x += delta_x)
      intensity0[i][j] = render_pixel (x, y, delta_x, delta_y, n);
  k += sqr (n);

#if 0
  do
    {
      sum = 0;
      deltamax = 0;
      n *= 2;
      for (j = 0, y = Y_MAX - j0 * delta_y; j < h; j++, y -= delta_y)
	for (i = 0, x = X_MIN + i0 * delta_x; i < w; i++, x += delta_x)
	  {
	    intensity1[i][j] = render_pixel (x, y, delta_x, delta_y, n);
	    delta = sqr (intensity0[i][j] - intensity1[i][j]);
	    intensity0[i][j] = (k * intensity0[i][j] +
				sqr (n) * intensity1[i][j]) / (k + sqr (n));
	    sum += delta;
	    deltamax = delta > deltamax ? delta : deltamax;
	  }
      k += sqr (n);
fprintf (stderr, "%d %d %f %f %d\n", i0, j0, sum / sqr (n), deltamax, k);
    }
  while (deltamax > .1); // && n <= 16);
#endif
  
  for (j = 0; j < h; j++)
    for (i = 0; i < w; i++)
      {
	I = intensity0[i][j];
	I = I < 0.0 ? 0.0 : I > 1.0 ? 1.0 : I;
	J = (int)(255 * I + .5);
	printf ("%c%c%c", J, J, J);
      }
  fflush (stdout);
}

void
render (int view_width, int view_height, int rays_per_pixel)
{
  int x, y;
  
  for (y = 0; y < view_height; y += 8)
    for (x = 0; x < view_width; x += 8)
      render_area (x, y, 8, 8, view_width, view_height, rays_per_pixel);
}
