/*

trace.c

(C) 1998 lars brinkhoff.

*/

#define TRACE_MAX_DEPTH 100

#include <math.h>

#include "ray.h"
#include "objects.h"
#include "lasr.h"
#include "intersection.h"

Real
trace (Ray ray, int i, int j, int n)
{
  Object *object;
  Real alpha;
  Vector normal;
  Real A;
  Ray ray2;
  Real a, b, x;
  Material mat;

  if (ray.depth == TRACE_MAX_DEPTH)
    {
//      fprintf (stderr, "%d recursive calls to trace ().\n", TRACE_MAX_DEPTH);
      return 0.0;
    }

  find_first_intersection (ray, &object, &alpha);
  
  if (object == 0)
    {
      if (sky.f != 0)
	return sky.f (&sky, &ray.d);
      else
	return 0;
    }
  
  ray2.p.x = ray.p.x + (alpha - epsilon) * ray.d.x;
  ray2.p.y = ray.p.y + (alpha - epsilon) * ray.d.y;
  ray2.p.z = ray.p.z + (alpha - epsilon) * ray.d.z;
  ray2.t = ray.t;
  ray2.depth = ray.depth + 1;

  object->normal (&normal, object, &ray2.p, ray.t);

  if (object->material->f)
    object->material->f (object, &ray2.p, &normal, &mat);
  else
    mat = *object->material;
  
  if (mat.emit > 1)
    return mat.emit;

  x = myrand (-mat.diff, mat.spec, 1);
  if (x < 0.0)
    {
      // Diffuse reflection.
      a = 2 * pi * i / n + myrand (-pi / n, pi / n, 1);
      b = pi - 2 * acos ((Real)j / n + myrand (-0.5 / n, 0.5 / n, 1));
      ray2.d.x = cos (a) * cos (b);
      ray2.d.y = sin (a) * cos (b);
      ray2.d.z =           sin (b);
      
      if (vector_scalar (ray2.d, normal) < 0.0)
	{
	  ray2.d.x = -ray2.d.x;
	  ray2.d.y = -ray2.d.y;
	  ray2.d.z = -ray2.d.z;
	}

      return mat.emit + (mat.diff + mat.spec) *
#if 1
	vector_scalar (ray2.d, normal) *
#endif
	trace (ray2, i, j, n);
    }
  else
    {
      // Specular reflection.
      A = 2.0 * vector_scalar (normal, ray.d);
      ray2.d.x = ray.d.x - A * normal.x + mat.grain * myrand (1, -1, 3);
      ray2.d.y = ray.d.y - A * normal.y + mat.grain * myrand (1, -1, 3);
      ray2.d.z = ray.d.z - A * normal.z + mat.grain * myrand (1, -1, 3);
      vector_normalize (ray2.d);

      return mat.emit + (mat.diff + mat.spec) * trace (ray2, i, j, n);
    }
}
