// lasrd/main.c
// (C) 1998 lars brinkhoff

#include <pwd.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/resource.h>

#define MAX_NUMBER_OF_JOBS 10

// Priority values.
#define HIGH_PRIORITY 10
#define LOW_PRIORITY 19

int number_of_jobs = 0;
pid_t job_pid[MAX_NUMBER_OF_JOBS];

void
reaper (int dummy)
{
  union wait status;
  pid_t pid;
  int i;
  
fprintf (stderr, "child has died\n");
  pid = wait3 (&status, 0, 0);
  if (pid == -1)
    return;

fprintf (stderr, "reaped child %d\n", pid);

  // Remove pid from the job table.
  for (i = 0; i < number_of_jobs; i++)
    if (job_pid[i] == pid)
      {
	for (i++; i < number_of_jobs; i++)
	  job_pid[i - 1] = job_pid[i];
      }
  number_of_jobs--;
      
  {
    int i;
    
    fprintf (stderr, "job list now:\n");
    for (i = 0; i < number_of_jobs; i++)
      fprintf (stderr, "  %d\n", job_pid[i]);
  }

  // Make sure the first job in line has a higher priority.
  if (number_of_jobs > 0)
    if (setpriority (PRIO_PROCESS, job_pid[0], HIGH_PRIORITY) == -1)
      {
	perror ("setpriority");
	if (geteuid () != 0)
	  fprintf (stderr, "Error possibly caused by not running as root.\n");
      }

  signal (SIGCHLD, reaper);
}

// Open a TCP/IP port with port number port, and listen to it.
int
server_socket (int port)
{
  struct sockaddr_in address;
  int S;

  S = socket (PF_INET, SOCK_STREAM, 0);
  if (S == -1)
    {
      perror ("socket");
      exit (1);
    }
  
  address.sin_family = PF_INET;
  address.sin_port = htons (port);
  address.sin_addr.s_addr = INADDR_ANY;
  
  if (bind (S, (struct sockaddr *)&address, sizeof (address)) == -1)
    {
      perror ("bind");
      exit (1);
    }

  if (listen (S, 5) == -1)
    {
      perror ("listen");
      exit (1);
    } 

  return S;
}

int
main (int argc, char *argv[])
{
  struct sockaddr_in address;
  int S, T;
  pid_t pid;
  int n;

  if (geteuid () != 0)
    fprintf (stderr,
	     "Warning: need to be run as root to be fully functional.\n");

  // Behave in a daemon-like manner.
// It will eventually.
//  if (fork ())
//    exit (0);

  signal (SIGCHLD, reaper);

  S = server_socket (10111);

  for (;;)
    {
      n = sizeof (address);
      T = accept (S, (struct sockaddr *)&address, &n);

      number_of_jobs++;

      if (number_of_jobs == MAX_NUMBER_OF_JOBS)
	{
	  fprintf (stderr, "Rejected connection from %08x: too many jobs.\n",
		   address.sin_addr.s_addr);
	  close (T);
	  continue;
	}

      fprintf (stderr, "Accepted connection from %08x.\n",
	       address.sin_addr.s_addr);

      pid = fork ();
      if (pid)
	{
	  job_pid[number_of_jobs - 1] = pid;
	  close (T);
	  {
	    int i;

	    fprintf (stderr, "job list now:\n");
	    for (i = 0; i < number_of_jobs; i++)
	      fprintf (stderr, "  %d\n", job_pid[i]);
	  }
	}
      else
	{
	  // If running as root, change to another user id.
	  if (geteuid () == 0)
	    {
	      struct passwd *passwd;

	      passwd = getpwnam ("nobody");
	      if  (passwd == 0)
		fprintf (stderr, "No such user: %s\n", "nobody");
	      else
		{
		  setuid (passwd->pw_uid);
		  seteuid (passwd->pw_uid);
		}
	    }
	  
	  if (number_of_jobs > 1)
	    nice (LOW_PRIORITY);
	  else
	    nice (HIGH_PRIORITY);

	  close (0);
	  dup (T);
	  close (1);
	  dup (T);
	  close (2);
	  dup (T);
	  close (T);
	  execle ("/home/lars/bin/lasr", "lasr", 0, 0);
	  exit (1);
	}
    }
}





























